# Copyright 2022 StorPool
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for the StorPool backend charm for Cinder"""

import unittest

import mock
from ops.model import ActiveStatus, BlockedStatus
from ops.testing import Harness

from src.charm import CinderCharmBase

BACKEND_NAME = "test"
STORPOOL_TEMPLATE = "test_template"
SP_API_HTTP_HOST = "127.0.0.1"
SP_API_HTTP_HOST_ALT = "127.1.0.1"
SP_AUTH_TOKEN = "1000000000000000000"
ISCSI_PORTAL_GROUP = "data-galore"


class TestCinderStorPoolCharm(unittest.TestCase):
    """
    Class that contains the StorPool charm unit tests
    """

    # pylint: disable=missing-function-docstring

    def setUp(self):
        self.harness = Harness(CinderCharmBase)
        self.addCleanup(self.harness.cleanup)
        self.harness.begin()
        self.harness.set_leader(True)
        backend = self.harness.add_relation("storage-backend", "cinder")
        # Bypass writing to /etc/storpool.conf
        with mock.patch("pathlib.Path.open", mock.mock_open()):
            self.harness.update_config(
                {
                    "volume-backend-name": BACKEND_NAME,
                    "storpool-template": STORPOOL_TEMPLATE,
                    "sp-api-http-host": SP_API_HTTP_HOST,
                    "sp-auth-token": SP_AUTH_TOKEN,
                    "iscsi-portal-group": ISCSI_PORTAL_GROUP,
                }
            )
        self.harness.add_relation_unit(backend, "cinder/0")

    def get_cinder_config(self):
        return dict(self.harness.charm.cinder_configuration(dict(self.harness.model.config)))

    def test_cinder_base(self):
        self.assertEqual(self.harness.framework.model.app.name, "cinder-storpool")
        # Test that charm is active upon installation.
        with mock.patch("pathlib.Path.open", mock.mock_open()):
            self.harness.update_config({})
        self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

        with mock.patch("pathlib.Path.open", mock.mock_open()):
            conf = self.get_cinder_config()
        self.assertDictEqual(
            conf,
            {
                "volume_driver": "cinder.volume.drivers.storpool.StorPoolDriver",
                "volume_backend_name": BACKEND_NAME,
                "storpool_template": STORPOOL_TEMPLATE,
                "sp_api_http_host": SP_API_HTTP_HOST,
                "sp_api_http_port": 81,
                "sp_auth_token": SP_AUTH_TOKEN,
                "iscsi_export_to": "*",
                "iscsi_portal_group": ISCSI_PORTAL_GROUP,
                "use_multipath_for_image_xfer": True,
                "enforce_multipath_for_image_xfer": True,
            },
        )

    def test_multipath_config(self):
        conf = None
        with mock.patch("pathlib.Path.open", mock.mock_open()):
            self.harness.update_config({"use-multipath": True})
            conf = self.get_cinder_config()
        self.assertEqual(conf["volume_backend_name"], BACKEND_NAME)
        self.assertTrue(conf.get("use_multipath_for_image_xfer"))
        self.assertTrue(conf.get("enforce_multipath_for_image_xfer"))

    def test_invalid_cinder_configurations(self):
        with mock.patch("pathlib.Path.open", mock.mock_open()):
            self.harness.update_config(unset=["storpool-template"])
            self.assertTrue(isinstance(self.harness.model.unit.status, BlockedStatus))

            self.harness.update_config({"storpool-template": STORPOOL_TEMPLATE})
            self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

            self.harness.update_config(unset=["sp-api-http-host"])
            self.assertTrue(isinstance(self.harness.model.unit.status, BlockedStatus))

            self.harness.update_config({"sp-api-http-host": SP_API_HTTP_HOST})
            self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

            self.harness.update_config(unset=["sp-auth-token"])
            self.assertTrue(isinstance(self.harness.model.unit.status, BlockedStatus))

            self.harness.update_config({"sp-auth-token": SP_AUTH_TOKEN})
            self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

            self.harness.update_config({"protocol": "block"})
            self.assertTrue(isinstance(self.harness.model.unit.status, BlockedStatus))

            self.harness.update_config({"protocol": "iscsi"})
            self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

            self.harness.update_config({"protocol": "something_invalid"})
            self.assertTrue(isinstance(self.harness.model.unit.status, BlockedStatus))

            self.harness.update_config({"protocol": "iscsi"})
            self.assertTrue(isinstance(self.harness.model.unit.status, ActiveStatus))

        # Cannot test for missing options with default values as it seems the
        # defaults are applied after update_config() below:
        # self.harness.update_config(unset=["protocol"])
        # self.harness.update_config(unset=["sp_api_http_port"])

    def test_storpool_conf_gets_updated(self):
        fake_open = mock.mock_open()

        with mock.patch("pathlib.Path.open", fake_open):
            self.harness.update_config({"sp-api-http-host": SP_API_HTTP_HOST_ALT})

        fake_open.assert_called_with(mode="w", encoding="UTF-8", errors=None)
        handle = fake_open()
        write_calls = [
            mock.call(
                "# Do not edit; this file is generated by the cinder-storpool charm.\n"
                + f"SP_API_HTTP_HOST={SP_API_HTTP_HOST_ALT}\n"
                + "SP_API_HTTP_PORT=81\n"
                + f"SP_AUTH_TOKEN={SP_AUTH_TOKEN}\n"
            ),
        ]
        handle.write.assert_has_calls(write_calls)
